/*
 * Copyright 2009-2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <algorithm>

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "configuration.hpp"
#include "geom.hpp"
#ifdef WITH_ZFS
#include "zpool.hpp"
#endif
#include "smtp.hpp"

using namespace std;

string pidFileName = "/var/run/geomWatch.pid";

void signalHandler(int signal) {
  switch (signal) {
    case SIGTERM:
      unlink(pidFileName.c_str());
      exit(0);
  }
}

const std::string pad(const std::string _string, size_t length) {
  if (_string.length() < length) {
    return _string + string(length - _string.length(), ' ');
  }
  return _string;
}

int main(int argc, char *argv[]) {
  char option, *hostname;
  ofstream pidFile;
  Configuration conf;
  vector <Geom> geoms;
#ifdef WITH_ZFS
  vector <ZPool> zpools;
  map <string, pair <vdev_state_t, string> >::const_iterator devItr, lastDevItr;
#endif
  SMTP smtp;
  string geom, configFileName = "/usr/local/etc/geomWatch.conf", crlf = "\r\n";
  size_t slash;
  bool lastUpdate;
  long hostnameLength = sysconf(_SC_HOST_NAME_MAX);
  ostringstream message;
  vector <string> lost;
  pid_t pid;
  signal(SIGTERM, signalHandler);
  while ((option = getopt(argc, argv, "c:p:")) != -1) {
    switch (option) {
      case 'c':
        configFileName = optarg;
        break;
      case 'p':
        pidFileName = optarg;
        break;
      default:
        return 1;
    }
  }
  if (!conf.open(configFileName)) {
    cerr << argv[0] << ": " << conf.error() << endl;
    return 1;
  }
#ifdef WITH_ZFS
  if (conf.getStrings("geom").size() == 0 &&
      conf.getStrings("zpool").size() == 0) {
    cerr << argv[0] << ": no geoms or ZFS zpools specified" << endl;
    return 1;
  }
#else
  if (conf.getStrings("geom").size() == 0) {
    cerr << argv[0] << ": no geoms specified" << endl;
    return 1;
  }
#endif
  if (conf.getStrings("recipient").size() == 0) {
    cerr << argv[0] << ": no recipients specified" << endl;
    return 1;
  }
  if (conf.getNumber("interval") == std::numeric_limits <size_t>::max()) {
    cerr << argv[0] << ": no interval specified" << endl;
    return 1;
  }
  if (conf.getNumber("smtpServer") == std::numeric_limits <size_t>::max()) {
    cerr << argv[0] << ": no SMTP server specified" << endl;
    return 1;
  }
  if (conf.getNumber("smtpAuth") != std::numeric_limits <size_t>::max()) {
    if (conf.getString("smtpUser") == "") {
      cerr << argv[0] << ": no SMTP user specified" << endl;
      return 1;
    }
    if (conf.getString("smtpPassword") == "") {
      cerr << argv[0] << ": no SMTP password specified" << endl;
      return 1;
    }
  }
  if (hostnameLength == -1) {
    cerr << argv[0] << ": sysconf(_SC_HOST_NAME_MAX): " << strerror(errno)
         << endl;
    return 1;
  }
  hostname = (char*)malloc(hostnameLength + 1);
  if (hostname == NULL) {
    cerr << argv[0] << ": malloc(" << hostnameLength << "): " << strerror(errno)
         << endl;
    return 1;
  }
  if (gethostname(hostname, hostnameLength + 1) == -1) {
    cerr << argv[0] << ": gethostname(): " << strerror(errno) << endl;
    return 1;
  }
  for (size_t i = 0; i < conf.getStrings("geom").size(); ++i) {
    geom = conf.getStrings("geom")[i];
    slash = geom.find('/');
    if (slash == string::npos) {
      cerr << argv[0] << ": no geom specified for class " << geom << endl;
      return 1;
    }
    geoms.push_back(Geom(geom.substr(0, slash), geom.substr(slash + 1)));
    if (!(geoms.rbegin() -> update())) {
      cerr << argv[0] << ": " << geoms.rbegin() -> error() << endl;
      return 1;
    }
  }
#ifdef WITH_ZFS
  for (size_t i = 0; i < conf.getStrings("zpool").size(); ++i) {
    zpools.push_back(ZPool(conf.getStrings("zpool")[i]));
    if (!(zpools.rbegin() -> update())) {
      cerr << argv[0] << ": " << zpools.rbegin() -> error() << endl;
      return 1;
    }
  }
#endif
  if (!smtp.initialize(conf.getString("smtpServer"), conf.getNumber("smtpAuth"),
                       conf.getString("smtpUser"),
                       conf.getString("smtpPassword"),
                       conf.getStrings("recipient"))) {
    cerr << argv[0] << ": " << smtp.error() << endl;
    return 1;
  }
  if (!smtp.setFrom(conf.getString("from"))) {
    cerr << argv[0] << ": " << smtp.error() << endl;
    return 1;
  }
  if (chdir("/") != 0) {
    cerr << argv[0] << ": chdir(): /: " << strerror(errno) << endl;
    return 1;
  }
  if (close(STDIN_FILENO) != 0) {
    cerr << argv[0] << ": close(): stdin: " << strerror(errno) << endl;
    return 1;
  }
  if (close(STDOUT_FILENO) != 0) {
    cerr << argv[0] << ": close(): stdout: " << strerror(errno) << endl;
    return 1;
  }
  pid = fork();
  if (pid < 0) {
    cerr << argv[0] << ": fork(): " << strerror(errno) << endl;
    return 1;
  }
  /* Exit if this is the parent process. */
  if (pid != 0) {
    return 0;
  }
  /*
   * Otherwise, write the daemonized child's PID to the PID file and enter the
   * main GEOM- and ZFS-checking loop.
   */
  pidFile.open(pidFileName.c_str());
  if (!pidFile) {
    cerr << argv[0] << ": open(): " << pidFileName << ": " << strerror(errno)
         << endl;
    return 1;
  }
  pidFile << getpid() << endl;
  pidFile.close();
  while (1) {
    sleep(conf.getNumber("interval"));
    for (size_t i = 0; i < geoms.size(); ++i) {
      lastUpdate = geoms[i].lastUpdate();
      if (!(geoms[i].update())) {
        if (lastUpdate == true) {
          smtp.setSubject("Problem checking GEOM " + geoms[i].name() + " on " +
                          hostname);
          message << "The following problem was encountered while checking " 
                  << "the status of GEOM " << geoms[i].name() << ": " << crlf
                  << crlf << '\t' << geoms[i].error() << crlf;
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
      else {
        if (geoms[i].providers().size() < geoms[i].lastProviders().size()) {
          smtp.setSubject("Component lost for GEOM " + geoms[i].name() +
                          " on " + hostname);
          if (geoms[i].status() == geoms[i].lastStatus()) {
            message << "The status of GEOM " << geoms[i].name() << " remains "
                    << "unchanged from " << geoms[i].status();
          }
          else {
            message << "The status of GEOM " << geoms[i].name() << " has "
                    << "changed from " << geoms[i].lastStatus() << " to "
                    << geoms[i].status();
          }
          lost.resize(geoms[i].lastProviders().size() -
                      geoms[i].providers().size());
          set_difference(geoms[i].lastProviders().begin(),
                         geoms[i].lastProviders().end(),
                         geoms[i].providers().begin(),
                         geoms[i].providers().end(),
                         lost.begin());
          message << ". The following components have been lost:" << crlf
                  << crlf;
          for (size_t j = 0; j < lost.size(); ++j) {
            message << '\t' << lost[j] << crlf;
          }
          message << crlf << "Remaining components:" << crlf << crlf;
          for (size_t j = 0; j < geoms[i].providers().size(); ++j) {
            message << '\t' << geoms[i].providers()[j] << crlf;
          }
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
    }
#ifdef WITH_ZFS
    for (size_t i = 0; i < zpools.size(); ++i) {
      lastUpdate = zpools[i].lastUpdate();
      if (!(zpools[i].update())) {
        if (lastUpdate == true) {
          smtp.setSubject("Problem checking ZFS pool " + zpools[i].name() +
                          " on " + hostname);
          message << "The following problem was encountered while checking "
                  << "the status of ZFS pool " << zpools[i].name() << ":"
                  << crlf << crlf << '\t' << zpools[i].error() << crlf;
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
      else {
        lost.clear();
        for (size_t j = 0; j < zpools[i].vdevs().size(); ++j) {
          if (zpools[i].vdevs()[j].devs().size() == 0) {
            if (zpools[i].vdevs()[j].status() != VDEV_STATE_HEALTHY &&
                zpools[i].lastVDevs()[j].status() == VDEV_STATE_HEALTHY) {
              lost.push_back(pad(zpools[i].vdevs()[j].name(), 20) +
                             zpools[i].vdevs()[j].statusName() + ')');
            }
          }
          else {
            for (devItr = zpools[i].vdevs()[j].devs().begin();
                 devItr != zpools[i].vdevs()[j].devs().end(); ++devItr) {
              lastDevItr = zpools[i].lastVDevs()[j].devs().find(devItr -> first);
              if (devItr -> second.first != VDEV_STATE_HEALTHY &&
                  lastDevItr != zpools[i].lastVDevs()[j].devs().end() &&
                  lastDevItr -> second.first == VDEV_STATE_HEALTHY) {
                lost.push_back(pad(devItr -> first, 20) +
                               devItr -> second.second);
              }
            }
          }
        }
        if (lost.size() > 0) {
          smtp.setSubject("Component lost for ZFS pool " + zpools[i].name() +
                          " on " + hostname);
          if (zpools[i].status() == zpools[i].lastStatus()) {
            message << "The status of ZFS pool " << zpools[i].name()
                    << " remains unchanged from " << zpools[i].statusName();
          }
          else {
            message << "The status of ZFS pool " << zpools[i].name() << " has "
                    << "changed from " << zpools[i].lastStatusName() << " to "
                    << zpools[i].statusName();
          }
          message << ". The following components have been lost:" << crlf
                  << crlf << '\t' << pad("Name", 20) << "State" << crlf << crlf;
          for (size_t j = 0; j < lost.size(); ++j) {
            message << '\t' << pad(lost[j], 20) << crlf;
          }
          message << crlf << "Resulting configuration:" << crlf << crlf
                  << '\t' << pad("Name", 20) << "State" << crlf << crlf
                  << '\t' << pad(zpools[i].name(), 20) << zpools[i].statusName()
                  << crlf;
          for (size_t j = 0; j < zpools[i].vdevs().size(); ++j) {
            message << "\t  " << pad(zpools[i].vdevs()[j].name(), 18)
                    << zpools[i].vdevs()[j].statusName() << crlf;
            for (devItr = zpools[i].vdevs()[j].devs().begin();
                 devItr != zpools[i].vdevs()[j].devs().end(); ++devItr) {
              message << "\t    " << pad(devItr -> first, 16)
                      << devItr -> second.second << crlf;
            }
          }
          smtp.setMessage(message.str());
          message.str("");
          if (!smtp.send()) {
            cerr << argv[0] << ": " << smtp.error() << endl;
          }
        }
      }
    }
#endif
  }
  /* Not reached. */
  return 0;
}
