/*
 * Copyright 2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ZPOOL_HPP
#define ZPOOL_HPP

#include <string>
#include <map>

#include <libzfs.h>
#include <errno.h>

#include "vdev.hpp"

class ZPool {
  public:
    ZPool(const std::string &name);
    const std::string &name() const;
    const std::string &error() const;
    bool update();
    const bool &lastUpdate() const;
    const vdev_state_t &status() const;
    const std::string &statusName() const;
    const vdev_state_t &lastStatus() const;
    const std::string &lastStatusName() const;
    const std::vector <VDev> &vdevs() const;
    const std::vector <VDev> &lastVDevs() const;
  private:
    std::string _name;
    libzfs_handle *zfs;
    zpool_handle *zpool;
    u_int c, numVDevs, numDevs;
    nvlist_t *poolConfig;
    nvlist_t *poolList;
    nvlist_t **vdevList;
    nvlist_t **devList;
    vdev_stat_t *stat;
    char *vdevName;
    vdev_state_t _status;
    std::string _statusName;
    vdev_state_t _lastStatus;
    std::string _lastStatusName;
    std::vector <VDev> _vdevs;
    std::vector <VDev> _lastVDevs;
    bool _lastUpdate;
    std::string errorMessage;
};

ZPool::ZPool(const std::string &name) {
  _lastUpdate = true;
  _name = name;
}

const std::string &ZPool::name() const {
  return _name;
}

const std::string &ZPool::error() const {
  return errorMessage;
}

bool ZPool::update() {
  zfs = libzfs_init();
  if (zfs == NULL) {
    _lastUpdate = false;
    errorMessage = "update(): libzfs_init(): failed to initialize ZFS library";
    return false;
  }
  zpool = zpool_open_canfail(zfs, _name.c_str());
  if (zpool == NULL) {
    _lastUpdate = false;
    errorMessage = "update(): zpool_open(): ";
    errorMessage += _name;
    errorMessage += ": pool not found";
    libzfs_fini(zfs);
    return false;
  }
  poolConfig = zpool_get_config(zpool, NULL);
  verify(nvlist_lookup_nvlist(poolConfig, ZPOOL_CONFIG_VDEV_TREE,
                              &poolList) == 0);
  verify(nvlist_lookup_uint64_array(poolList, ZPOOL_CONFIG_STATS,
         (uint64_t**)&stat, &c) == 0);
  _lastStatus = _status;
  _lastStatusName = _statusName;
  _status = (vdev_state_t)(stat -> vs_state);
  _statusName = zpool_state_to_name((vdev_state_t)(stat -> vs_state),
                                    (vdev_aux_t)(stat -> vs_aux));
  if (nvlist_lookup_nvlist_array(poolList, ZPOOL_CONFIG_CHILDREN, &vdevList,
                                 &numVDevs) != 0) {
    numVDevs = 0;
  }
  _lastVDevs = _vdevs;
  _vdevs.clear();
  for (uint_t i = 0; i < numVDevs; ++i) {
    verify(nvlist_lookup_uint64_array(vdevList[i], ZPOOL_CONFIG_STATS,
                                      (uint64_t**)&stat, &c) == 0);
    vdevName = zpool_vdev_name(zfs, zpool, vdevList[i]);
    if (vdevName == NULL) {
      errorMessage = "update(): zpool_vdev_name(): ";
      errorMessage += strerror(errno);
      nvlist_free(poolList);
      zpool_close(zpool);
      libzfs_fini(zfs);
      _lastUpdate = false;
      return false;
    }
    _vdevs.push_back(VDev(vdevName, (vdev_state_t)(stat -> vs_state),
                          zpool_state_to_name((vdev_state_t)(stat -> vs_state),
                                              (vdev_aux_t)(stat -> vs_aux))));
    free(vdevName);
    if (nvlist_lookup_nvlist_array(vdevList[i], ZPOOL_CONFIG_CHILDREN,
                                   &devList, &numDevs) != 0) {
      numDevs = 0;
    }
    for (uint_t j = 0; j < numDevs; ++j) {
      verify(nvlist_lookup_uint64_array(devList[j], ZPOOL_CONFIG_STATS,
                                        (uint64_t**)&stat, &c) == 0);
      vdevName = zpool_vdev_name(zfs, zpool, devList[j]);
      if (vdevName == NULL) {
        errorMessage = "update(): zpool_vdev_name(): ";
        errorMessage += strerror(errno);
        nvlist_free(poolList);
        zpool_close(zpool);
        libzfs_fini(zfs);
        _lastUpdate = false;
        return false;
      }
      _vdevs.rbegin() -> add(vdevName, (vdev_state_t)(stat -> vs_state),
                             zpool_state_to_name((vdev_state_t)(stat -> vs_state),
                                                 (vdev_aux_t)(stat -> vs_aux)));
      free(vdevName);
    }
  }
  nvlist_free(poolList);
  zpool_close(zpool);
  libzfs_fini(zfs);
  _lastUpdate = true;
  return true;
}

const bool &ZPool::lastUpdate() const {
  return _lastUpdate;
}

const vdev_state_t &ZPool::status() const {
  return _status;
}

const std::string &ZPool::statusName() const {
  return _statusName;
}

const vdev_state_t &ZPool::lastStatus() const {
  return _lastStatus;
}

const std::string &ZPool::lastStatusName() const {
  return _lastStatusName;
}

const std::vector <VDev> &ZPool::vdevs() const {
  return _vdevs;
}

const std::vector <VDev> &ZPool::lastVDevs() const {
  return _lastVDevs;
}

#endif
